import numpy as np 
from scipy.ndimage import distance_transform_edt
from . import utilities as util 
from math import factorial

def shift(xs, n, m, fill_value=np.nan):
    # shifts a 2D array xs by n rows, m columns and fills the new region with fill_value

    e = xs.copy()
    if n!=0:
        if n >= 0:
            e[:n,:] = fill_value
            e[n:,:] =  e[:-n,:]
        else:
            e[n:,:] = fill_value
            e[:n,:] =  e[-n:,:]
   
       
    if m!=0:
        if m >= 0:
            e[:,:m] = fill_value
            e[:,m:] =  e[:,:-m]
        else:
            e[:,m:] = fill_value
            e[:,:m] =  e[:,-m:]
    return e

def construct_command_basis( basis='Zernike_pinned_edges', number_of_modes = 20, Nx_act_DM = 12, Nx_act_basis = 12, act_offset=(0,0), without_piston=True):
    """
    returns a change of basis matrix M2C to go from modes to DM commands, where columns are the DM command for a given modal basis. e.g. M2C @ [0,1,0,...] would return the DM command for tip on a Zernike basis. Modes are normalized on command space such that <M>=0, <M|M>=1. Therefore these should be added to a flat DM reference if being applied.    

    basis = string of basis to use
    number_of_modes = int, number of modes to create
    Nx_act_DM = int, number of actuators across DM diameter
    Nx_act_basis = int, number of actuators across the active basis diameter
    act_offset = tuple, (actuator row offset, actuator column offset) to offset the basis on DM (i.e. we can have a non-centered basis)
    IM_covariance = None or an interaction matrix from command to measurement space. This only needs to be provided if you want KL modes, for this the number of modes is infered by the shape of the IM matrix. 
     
    """

    
   
    # shorter notations
    #Nx_act = DM.num_actuators_width() # number of actuators across diameter of DM.
    #Nx_act_basis = actuators_across_diam
    c = act_offset
    # DM BMC-3.5 is 12x12 missing corners so 140 actuators , we note down corner indicies of flattened 12x12 array.
    corner_indices = [0, Nx_act_DM-1, Nx_act_DM * (Nx_act_DM-1), -1]

    

    if basis == 'Hadamard':
        # BMC multi-3.5 DM is 12x12 with missing corners - not multiple of 2 so cannot apply standard method of construction
        # first 
        # Define a known 12x12 Hadamard matrix
        H_12 = np.array([
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            [1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1],
            [1, 1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1],
            [1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1, 1],
            [1, 1, 1, 1, -1, -1, -1, -1, 1, 1, 1, 1],
            [1, -1, 1, -1, -1, 1, -1, 1, 1, -1, 1, -1],
            [1, 1, -1, -1, -1, -1, 1, 1, 1, 1, -1, -1],
            [1, -1, -1, 1, -1, 1, 1, -1, 1, -1, -1, 1],
            [1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1],
            [1, -1, 1, -1, 1, -1, -1, 1, -1, 1, -1, 1],
            [1, 1, -1, -1, 1, 1, 1, 1, -1, -1, -1, -1],
            [1, -1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1]
        ])

        # Generate a 144x144 Hadamard matrix using Kronecker product
        H_144 = np.kron(H_12, H_12)

        # Define the mask to remove corners without affecting the 2D structure
        corner_mask = np.ones((12, 12), dtype=bool)
        corner_mask[0, 0] = False  # top-left corner
        corner_mask[0, -1] = False  # top-right corner
        corner_mask[-1, 0] = False  # bottom-left corner
        corner_mask[-1, -1] = False  # bottom-right corner

        H_140 = []
        for h in H_144:
            H_140.append( h.reshape(12,12)[corner_mask] )
            
        if without_piston:
            M2C = np.array( H_140 )[1:].T #remove piston mode
        else:
            M2C = np.array( H_140 ).T # take transpose to make columns the modes in command space.


    # to deal with
    elif basis == 'Zernike':
        bmcdm_basis_list = []
        if without_piston:
            number_of_modes += 1 # we add one more mode since we dont include piston 

        raw_basis = zernike_basis(nterms=number_of_modes, npix=Nx_act_basis )
        for i,B in enumerate(raw_basis):
            # normalize <B|B>=1, <B>=0 (so it is an offset from flat DM shape)
            Bnorm = np.sqrt( 1/np.nansum( B**2 ) ) * B
            # pad with zeros to fit DM square shape and shift pixels as required to center
            # we also shift the basis center with respect to DM if required
            if np.mod( Nx_act_basis, 2) == 0:
                pad_width = (Nx_act_DM - B.shape[0] )//2
                padded_B = shift( np.pad( Bnorm , pad_width , constant_values=(np.nan,)) , c[0], c[1])
            else:
                pad_width = (Nx_act_DM - B.shape[0] )//2 + 1
                padded_B = shift( np.pad( Bnorm , pad_width , constant_values=(np.nan,)) , c[0], c[1])[:-1,:-1]  # we take off end due to odd numebr

            flat_B = padded_B.reshape(-1) # flatten basis so we can put it in the accepted DM command format
            np.nan_to_num(flat_B,0 ) # convert nan -> 0
            flat_B[corner_indices] = np.nan # convert DM corners to nan (so lenght flat_B = 140 which corresponds to BMC-3.5 DM)

            # now append our basis function removing corners (nan values)
            bmcdm_basis_list.append( flat_B[np.isfinite(flat_B)] )

        # our mode 2 command matrix
        if without_piston:
            M2C = np.array( bmcdm_basis_list )[1:].T #remove piston mode
        else:
            M2C = np.array( bmcdm_basis_list ).T # take transpose to make columns the modes in command space.


    elif basis == 'Zernike_pinned_edges':
        """
        designed for BMC multi 3.5 DM, define zernike basis on 10x10 central grid and 
        interpolate outside of this grid, pinning the value of perimeter actuators to the
        inner perimeter value. 
        """
        nact_len = 12 # must be 12
        # alway construct with piston cause we use as a filter, we delete piston later if specified by user
        b0 = construct_command_basis( basis='Zernike', number_of_modes = number_of_modes, Nx_act_DM = nact_len, Nx_act_basis = nact_len, act_offset=(0,0), without_piston=False)

        # put values outside pupil to nan 
        btmp = np.array( [util.get_DM_command_in_2D( bb ) for bb in b0.T])

        # interpolate
        nan_mask = btmp[0] #util.get_DM_command_in_2D( b0.T[0] != 0 )
        nan_mask[nan_mask==0] = np.nan

        #nan_mask = np.isnan(nan_mask)
        nearest_index = distance_transform_edt(np.isnan(nan_mask), return_distances=False, return_indices=True)

        # Use the indices to replace NaNs with the nearest non-NaN values

        with_corners = np.array( [ (nan_mask * bb)[tuple(nearest_index)] for bb in btmp[:]] ).T
        #filled_data = util.get_DM_command_in_2D( new_flat )[tuple(nearest_index)]


        # Define the indices of the corners to be removed
        corners = [(0, 0), (0, nact_len-1), (nact_len-1, 0), (nact_len-1, nact_len-1)]
        # Convert 2D corner indices to 1D
        corner_indices = [i * 12 + j for i, j in corners]

        # Flatten the array
        bmcdm_basis_list = []
        for w in with_corners.T:
            flattened_array = w.flatten()
            filtered_array = np.delete(flattened_array, corner_indices)

            bmcdm_basis_list.append( filtered_array )

        # piston was taken care of in construction of original zernike basis  b0 = construct_command_basis(
        if without_piston:
            control_basis = [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in bmcdm_basis_list[1:]]
        else:
            control_basis = [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in bmcdm_basis_list[:]]
        M2C = np.array( control_basis ).T

             
    elif basis == 'fourier':
        # NOT TESTED YET ON REAL DM!! 
        if without_piston:
            number_of_modes += 1 # we add one more mode since we dont include piston 

        # NOTE BECAUSE WE HAVE N,M DIMENSIONS WE NEED TO ROUND UP TO SQUARE NUMBER THE MIGHT NOT = EXACTLY number_of_modes
        n = round( number_of_modes**0.5 ) + 1 # number of modes = (n-1)*(m-1) , n=m => (n-1)**2 
        control_basis_dict  = develop_Fourier_basis( n, n ,P = 2 * Nx_act_DM, Nx = Nx_act_DM, Ny = Nx_act_DM )
        
        # create raw basis as ordered list from our dictionary
        raw_basis = []
        for i in range( n-1 ):
            for j in np.arange( i , n-1 ):
                if i==j:
                    raw_basis.append( control_basis_dict[i,i] )
                else:
                    raw_basis.append( control_basis_dict[i,j] ) # get either side of diagonal 
                    raw_basis.append( control_basis_dict[j,i] )
                    
        
        bmcdm_basis_list = []
        for i,B in enumerate(raw_basis):
            B = B.reshape(-1)
            B[corner_indices] = np.nan
            bmcdm_basis_list.append( B[np.isfinite(B)] )
        # flatten & normalize each basis cmd 
        # <M|M> = 1
        if without_piston:
            control_basis = [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in bmcdm_basis_list[1:]] #remove piston mode
        else:
            control_basis = [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in bmcdm_basis_list]# take transpose to make columns the modes in command space.
        M2C = np.array( control_basis ).T 

    elif basis == 'fourier_pinned_edges':
        """
        designed for BMC multi 3.5 DM, define zernike basis on 10x10 central grid and 
        interpolate outside of this grid, pinning the value of perimeter actuators to the
        inner perimeter value. 
        """
        n = round( number_of_modes**0.5 ) + 1 # number of modes = (n-1)*(m-1) , n=m => (n-1)**2 
        actlen_tmp = 10 # must be 10 for this option! we then calculate perimeter values here! 
        control_basis_dict  = develop_Fourier_basis( n, n ,P = 2 * actlen_tmp, Nx = actlen_tmp, Ny = actlen_tmp )
                
        # create raw basis as ordered list from our dictionary
        raw_basis = []
        for i in range( n-1 ):
            for j in np.arange( i , n-1 ):
                if i==j:
                    raw_basis.append( control_basis_dict[i,i] )
                else:
                    raw_basis.append( control_basis_dict[i,j] ) # get either side of diagonal 
                    raw_basis.append( control_basis_dict[j,i] )
                    
        # pin_outer_actuators_to_inner requires 10x10 input!!! creates 12x12 with missing corner pinning outer actuators 
        bmcdm_basis_list = np.array( [pin_outer_actuators_to_inner_diameter(bb.reshape(-1)) for bb in np.array( raw_basis)] )

        # <M|M> = 1
        if without_piston:
            control_basis = [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in bmcdm_basis_list[1:]] #remove piston mode
        else:
            control_basis = [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in bmcdm_basis_list]# take transpose to make columns the modes in command space.
        
        M2C = np.array( control_basis ).T 


    elif basis == 'Zonal': 
        #hardcoded for BMC multi3.5 DM (140 actuators)
        M2C = np.eye( 140 ) # we just consider this over all actuators (so defaults to 140 modes) 
        # we filter zonal basis in the eigenvectors of the control matrix. 
    
    elif basis == 'Zonal_pinned_edges':
        # pin edges of actuator
        b = np.eye(100) #
        bmcdm_basis_list = np.array( [pin_outer_actuators_to_inner_diameter(bb) for bb in b.T] )
        # <M|M> = 1
        control_basis = np.array( [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in bmcdm_basis_list] )

        M2C = np.array( control_basis ).T

    else:
        raise TypeError( ' input basis name invalid. Try: "Zonal", "Zonal_pinned_edges", "Zernike", "Zernike_pinned_edges", "fourier", "fourier_pinned_edges"  etc ')
    
    
    return(M2C)


def get_tip_tilt_vectors( dm_model='bmc_multi3.5' ,nact_len=12):
    tip = np.array([[n for n in np.linspace(-1,1,nact_len)] for _ in range(nact_len)])
    tilt = tip.T
    if dm_model == 'bmc_multi3.5':
        # Define the indices of the corners to be removed
        corners = [(0, 0), (0, nact_len-1), (nact_len-1, 0), (nact_len-1, nact_len-1)]
        # Convert 2D corner indices to 1D
        corner_indices = [i * 12 + j for i, j in corners]

        # remove corners
        tip_tilt_list = []
        for i,B in enumerate([tip,tilt]):
            B = B.reshape(-1)
            B[corner_indices] = np.nan
            tip_tilt_list.append( B[np.isfinite(B)] )
        
        tip_tilt = np.array( [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in tip_tilt_list] ).T

    else:
        tip_tilt = np.array( [np.sqrt( 1/np.nansum( cb**2 ) ) * cb.reshape(-1) for cb in [tip.reshape(-1),tilt.reshape(-1)]] ).T

    return( tip_tilt ) 


def fourier_vector(n, m, P = 2*12, Nx = 12, Ny = 12):
    """
    OR we can do it with complex exponetial, in-quadrature is real part, out of quadrature is imaginary 
    Normalized <Bx|Bx>=1 , <By|By>=1

    Parameters
    ----------
    n : TYPE
        DESCRIPTION.
    m : TYPE
        DESCRIPTION.
    P : TYPE, optional
        DESCRIPTION. The default is 2*12.
    Nx : TYPE, optional
        DESCRIPTION. The default is 12.
    Ny : TYPE, optional
        DESCRIPTION. The default is 12.

    Returns
    -------
    None.

    """
    x = np.linspace(-6,6,Nx)
    y = np.linspace(-6,6,Ny)
    X,Y = np.meshgrid(x,y)
    
    
    Bx = np.exp( 1J * 2 * np.pi * n/P * X )
    if np.sum( abs(Bx) ):
        Bx *= 1/np.sum( abs(Bx)**2 )**0.5

    By = np.exp( 1J * 2 * np.pi * m/P * Y )
    if np.sum( abs(By) ):
        By *= 1/np.sum( abs(By)**2 )**0.5
    
    return( Bx, By )

def develop_Fourier_basis( n,m ,P = 2*12, Nx = 12, Ny = 12):
    """
    

    Parameters
    ----------
    n : TYPE int
        DESCRIPTION. what order in x (column) dimension do we create Fourier basis for?
    m : TYPE int
        DESCRIPTION. what order in y (row) dimension do we create Fourier basis for?

    Returns
    -------
    basis_dict - a dictionary indexed by mode order tuple (n,m) with corresponding 2D Fourier basis
    
    
    # what is logical indexing? 
    basis naturally forms 2 NxN squares, one square corresponding to odd components (sin) in x,y other to even (cos)

    for each axis dimension cnt, with even numbers corresponding to even functions (cos), odd numbers to odd functions (sin)
    therefore to recover cos or sin order we simply divide by 2 and round (//2)

    we do not count piston  
    e.g. indexing for x dimension:
    0 = np.real(F_basis_x[0] )  
    1 = np.imag(F_basis_x[0] )  
    2 = np.iamg(F_basis_x[1] ) 
    3 = np.real(F_basis_x[1] ) 

    therefore for example index (3,2)
    B_(3,2) = np.real(F_basis_x[1] ) * np.imag(F_basis_y[1] )
    first index corresponds to variation across columns (x), 
    second index corresponds to variation across rows (y)

    """
    basis_dict = {}

    for x_idx in range(0,n):
        for y_idx in range(0,m):            
            #
            x_order = x_idx//2
            y_order = y_idx//2
            
            if not ((x_idx==0) | (y_idx==0)): # otherwise we get lots of pistons 
                Bx, By = fourier_vector(x_order, y_order, P , Nx , Ny )
                    
                if not np.mod(x_idx,2): #odd number take imaginary (odd) part
                    Bx_q = np.imag( Bx )
    
                else: # even number take real (even) part
                    Bx_q = np.real( Bx )
    
                    
                if not np.mod(y_idx,2): #odd number take imaginary (odd) part
                    By_q = np.imag( By )
                    
                else: # even number take real (even) part
                    By_q = np.real( By )
            
                #if x_idx > 1:
                mode_tmp = Bx_q * By_q - np.mean(Bx_q * By_q)
                if np.sum( mode_tmp**2):
                    mode_tmp *= 1/np.sum( mode_tmp**2)**0.5 #normalized <M|M>=1
                basis_dict[(x_idx-1,y_idx-1)] =  mode_tmp


    return(basis_dict)






def noll_indices(j):
    """
    copied from taken from pyzelda (2024) 
    # https://github.com/avigan/pyZELDA/blob/master/pyzelda/utils/zernike.py
    
    Convert from 1-D to 2-D indexing for Zernikes or Hexikes.

    Parameters
    ----------
    j : int
        Zernike function ordinate, following the convention of Noll et al. JOSA 1976.
        Starts at 1.

    """

    if j < 1:
        raise ValueError("Zernike index j must be a positive integer.")

    # from i, compute m and n
    # I'm not sure if there is an easier/cleaner algorithm or not.
    # This seems semi-complicated to me...

    # figure out which row of the triangle we're in (easy):
    n = int(np.ceil((-1 + np.sqrt(1 + 8 * j)) / 2) - 1)
    if n == 0:
        m = 0
    else:
        nprev = (n + 1) * (n + 2) / 2  # figure out which entry in the row (harder)
        # The rule is that the even Z obtain even indices j, the odd Z odd indices j.
        # Within a given n, lower values of m obtain lower j.

        resid = int(j - nprev - 1)

        if _is_odd(j):
            sign = -1
        else:
            sign = 1

        if _is_odd(n):
            row_m = [1, 1]
        else:
            row_m = [0]

        for i in range(int(np.floor(n / 2.))):
            row_m.append(row_m[-1] + 2)
            row_m.append(row_m[-1])

        m = row_m[resid] * sign

    return n, m

def _is_odd(integer):
    """Helper for testing if an integer is odd by bitwise & with 1."""
    return integer & 1

def zernike_basis(nterms=15, npix=512, rho=None, theta=None, **kwargs):
    """
    copied from taken from pyzelda (2024) 
    # https://github.com/avigan/pyZELDA/blob/master/pyzelda/utils/zernike.py
    
    Return a cube of Zernike terms from 1 to N each as a 2D array
    showing the value at each point. (Regions outside the unit circle on which
    the Zernike is defined are initialized to zero.)

    Parameters
    -----------
    nterms : int, optional
        Number of Zernike terms to return, starting from piston.
        (e.g. ``nterms=1`` would return only the Zernike piston term.)
        Default is 15.
    npix: int
        Desired pixel diameter for circular pupil. Only used if `rho`
        and `theta` are not provided.
    rho, theta : array_like
        Image plane coordinates. `rho` should be 0 at the origin
        and 1.0 at the edge of the circular pupil. `theta` should be
        the angle in radians.

    Other parameters are passed through to `poppy.zernike.zernike`
    and are documented there.
    """
    if rho is not None and theta is not None:
        # both are required, but validated in zernike1
        shape = rho.shape
        use_polar = True
    elif (theta is None and rho is not None) or (theta is not None and rho is None):
        raise ValueError("If you provide either the `theta` or `rho` input array, you must "
                             "provide both of them.")

    else:
        shape = (npix, npix)
        use_polar = False

    zern_output = np.zeros((nterms,) + shape)

    if use_polar:
        for j in range(nterms):
            zern_output[j] = zernike1(j + 1, rho=rho, theta=theta, **kwargs)
    else:
        for j in range(nterms):
            zern_output[j] = zernike1(j + 1, npix=npix, **kwargs)
    return zern_output



def zernike1(j, **kwargs):
    """ 
    copied from taken from pyzelda (2024) 
    # https://github.com/avigan/pyZELDA/blob/master/pyzelda/utils/zernike.py
    
    
    Return the Zernike polynomial Z_j for pupil points {r,theta}.

    For this function the desired Zernike is specified by a single index j.
    See zernike for an equivalent function in which the polynomials are
    ordered by two parameters m and n.

    Note that there are multiple contradictory conventions for labeling Zernikes
    with one single index. We follow that of Noll et al. JOSA 1976.

    Parameters
    ----------
    j : int
        Zernike function ordinate, following the convention of
        Noll et al. JOSA 1976

    Additional arguments are defined as in `poppy.zernike.zernike`.

    Returns
    -------
    zern : 2D numpy array
        Z_j evaluated at each (rho, theta)
    """
    n, m = noll_indices(j)
    return zernike(n, m, **kwargs)


def R(n, m, rho):
    """
    copied from taken from pyzelda (2024) 
    # https://github.com/avigan/pyZELDA/blob/master/pyzelda/utils/zernike.py
    
    Compute R[n, m], the Zernike radial polynomial

    Parameters
    ----------
    n, m : int
        Zernike function degree
    rho : array
        Image plane radial coordinates. `rho` should be 1 at the desired pixel radius of the
        unit circle
    """

    m = int(np.abs(m))
    n = int(np.abs(n))
    output = np.zeros(rho.shape)
    if _is_odd(n - m):
        return 0
    else:
        for k in range(int((n - m) / 2) + 1):
            coef = ((-1) ** k * factorial(round( n - k)) /
                    (factorial(k) * factorial( round( (n + m) / 2. - k) ) * factorial( round( (n - m) / 2. - k))) )
            output += coef * rho ** (n - 2 * k)
        return output


def zernike(n, m, npix=100, rho=None, theta=None, outside=np.nan,
            noll_normalize=True):
    """
    copied from taken from pyzelda (2024) 
    # https://github.com/avigan/pyZELDA/blob/master/pyzelda/utils/zernike.py
    
    Return the Zernike polynomial Z[m,n] for a given pupil.

    For this function the desired Zernike is specified by 2 indices m and n.
    See zernike1 for an equivalent function in which the polynomials are
    ordered by a single index.

    You may specify the pupil in one of two ways:
     zernike(n, m, npix)       where npix specifies a pupil diameter in pixels.
                               The returned pupil will be a circular aperture
                               with this diameter, embedded in a square array
                               of size npix*npix.
     zernike(n, m, rho=r, theta=theta)    Which explicitly provides the desired pupil coordinates
                               as arrays r and theta. These need not be regular or contiguous.

    The expressions for the Zernike terms follow the normalization convention
    of Noll et al. JOSA 1976 unless the `noll_normalize` argument is False.

    Parameters
    ----------
    n, m : int
        Zernike function degree
    npix: int
        Desired diameter for circular pupil. Only used if `rho` and
        `theta` are not provided.
    rho, theta : array_like
        Image plane coordinates. `rho` should be 0 at the origin
        and 1.0 at the edge of the circular pupil. `theta` should be
        the angle in radians.
    outside : float
        Value for pixels outside the circular aperture (rho > 1).
        Default is `np.nan`, but you may also find it useful for this to
        be 0.0 sometimes.
    noll_normalize : bool
        As defined in Noll et al. JOSA 1976, the Zernike definition is
        modified such that the integral of Z[n, m] * Z[n, m] over the
        unit disk is pi exactly. To omit the normalization constant,
        set this to False. Default is True.

    Returns
    -------
    zern : 2D numpy array
        Z(m,n) evaluated at each (rho, theta)
    """
    if not n >= m:
        raise ValueError("Zernike index m must be >= index n")
    if (n - m) % 2 != 0:
        print("Radial polynomial is zero for these inputs: m={}, n={} "
                  "(are you sure you wanted this Zernike?)".format(m, n))

    if theta is None and rho is None:
        x = (np.arange(npix, dtype=np.float64) - (npix - 1) / 2.) / ((npix - 1) / 2.)
        y = x
        xx, yy = np.meshgrid(x, y)

        rho = np.sqrt(xx ** 2 + yy ** 2)
        theta = np.arctan2(yy, xx)
    elif (theta is None and rho is not None) or (theta is not None and rho is None):
        raise ValueError("If you provide either the `theta` or `rho` input array, you must "
                             "provide both of them.")

    if not np.all(rho.shape == theta.shape):
        raise ValueError('The rho and theta arrays do not have consistent shape.')

    aperture = np.ones(rho.shape)
    aperture[np.where(rho > 1)] = 0.0  # this is the aperture mask

    if m == 0:
        if n == 0:
            zernike_result = aperture
        else:
            norm_coeff = np.sqrt(n + 1) if noll_normalize else 1
            zernike_result = norm_coeff * R(n, m, rho) * aperture
    elif m > 0:
        norm_coeff = np.sqrt(2) * np.sqrt(n + 1) if noll_normalize else 1
        zernike_result = norm_coeff * R(n, m, rho) * np.cos(np.abs(m) * theta) * aperture
    else:
        norm_coeff = np.sqrt(2) * np.sqrt(n + 1) if noll_normalize else 1
        zernike_result = norm_coeff * R(n, m, rho) * np.sin(np.abs(m) * theta) * aperture

    zernike_result[np.where(rho > 1)] = outside
    return zernike_result





def pin_outer_actuators_to_inner_diameter(inner_command):
    """
    input a basis defined on 10x10 grid and this will convert it to a
    12x12 grid without corners (BMC multi3.5 DM geometry) with the outer
    perimeter actuators pinned to the inner perimeter value
    """
    if len(inner_command) != 100:
        raise ValueError("Input command must be of length 100")

    inner_command = np.array(inner_command).reshape(10, 10)
    
    # Initialize a 12x12 grid with zeros
    command_140 = np.zeros((12, 12))

    # Map the inner 10x10 command to the corresponding position in the 12x12 grid
    command_140[1:11, 1:11] = inner_command

    # Set the perimeter actuators equal to the inner adjacent values
    # Top and bottom rows
    command_140[0, 1:11] = command_140[1, 1:11]
    command_140[11, 1:11] = command_140[10, 1:11]

    # Left and right columns
    command_140[1:11, 0] = command_140[1:11, 1]
    command_140[1:11, 11] = command_140[1:11, 10]

    # Corners (set these to zero since they are missing actuators)
    corners = [(0, 0), (0, 11), (11, 0), (11, 11)]
    for corner in corners:
        command_140[corner] = 0

    # Flatten the 12x12 grid to get the final 140-length command
    command_140_flat = command_140.flatten()

    # Remove the corner actuators (i.e., elements 0, 11, 132, 143)
    indices_to_remove = [0, 11, 132, 143]
    command_140_flat = np.delete(command_140_flat, indices_to_remove)

    return command_140_flat.tolist()

def pin_to_nearest_registered_with_missing_corners(dm_shape, missing_corners, registered_indices):
    """
    Pins non-registered actuators to the closest registered actuator, excluding missing corners.

    Parameters:
    - dm_shape: Tuple (rows, cols) representing the DM grid, e.g., (12, 12).
    - missing_corners: List of indices (in the flattened array) of missing corners.
    - registered_indices: 1D array of indices corresponding to actuators with registered values.

    Returns:
    - basis: 2D array (dm_shape[0] * dm_shape[1] - len(missing_corners), len(registered_indices))
             where each non-registered actuator is pinned to its closest registered actuator.
    """
    # Create the full DM grid with flattened indices
    flattened_size = dm_shape[0] * dm_shape[1]
    
    # Generate 2D coordinates for each point on the grid
    grid_coords = np.array(np.unravel_index(np.arange(flattened_size), dm_shape)).T
    
    # Remove missing corners from the grid and flatten the remaining actuators
    valid_indices = np.setdiff1d(np.arange(flattened_size), missing_corners)
    valid_coords = grid_coords[valid_indices]

    # Extract coordinates of the registered actuators
    registered_coords = grid_coords[registered_indices]
    
    # Initialize the basis matrix for valid actuators
    basis = np.zeros((len(valid_indices), len(registered_indices)))
    
    # For each valid actuator, find the closest registered actuator
    for idx, valid_idx in enumerate(valid_indices):
        if valid_idx in registered_indices:
            # If the actuator is registered, set its basis vector to be identity
            basis[idx, registered_indices == valid_idx] = 1.0
        else:
            # If the actuator is not registered, pin it to the nearest registered actuator
            distances = distance.cdist([grid_coords[valid_idx]], registered_coords)
            nearest_idx = np.argmin(distances)
            # Pin to the nearest registered actuator
            basis[idx, nearest_idx] = 1.0
    
    #<m|m>=1
    basis_norm = np.array( [b/np.sum(b**2)**0.5 for b in basis.T] ).T
    
    
    return basis_norm








import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np

import pyzelda.zelda as zelda
import pyzelda.ztools as ztools
import pyzelda.utils.aperture as aperture
import pyzelda.utils.zernike as zernike
import pyzelda.utils.mft as mft

import numpy as np
import matplotlib.pyplot as plt
from types import SimpleNamespace
import importlib 
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid1 import make_axes_locatable
import os 
import sys
#import aotools

def add_project_root_to_sys_path(project_root_name="BaldrApp"):
    """
    Adds the project root directory to sys.path to allow importing from shared modules.
    
    Args:
        project_root_name (str): The name of the project root directory.
    """
    try:
        # Attempt to use __file__ to get the current script's directory
        current_dir = os.path.dirname(os.path.abspath(__file__))
    except NameError:
        # Fallback to current working directory (useful in interactive environments)
        current_dir = os.getcwd()

    # Traverse up the directory tree to find the project root
    project_root = current_dir
    while True:
        if os.path.basename(project_root) == project_root_name:
            break
        new_project_root = os.path.dirname(project_root)
        if new_project_root == project_root:
            # Reached the filesystem root without finding the project root
            project_root = None
            break
        project_root = new_project_root

    if project_root and project_root not in sys.path:
        sys.path.append(project_root)
        print(f"Added '{project_root}' to sys.path")
    elif project_root is None:
        print(f"Error: '{project_root_name}' directory not found in the directory hierarchy.")
    else:
        print(f"'{project_root}' is already in sys.path")

# Call the function to add the project root
add_project_root_to_sys_path()
from baldrapp.common import baldr_core as bldr
from baldrapp.common import DM_basis as gen_basis
from baldrapp.common import utilities as util
from baldrapp.common import DM_registration as DM_reg
from baldrapp.common import phasescreens
from baldrapp.common import pupils


class RecursiveNamespaceToClass:
    def __init__(self, namespace):
        # Iterate through all attributes of the SimpleNamespace
        for key, value in vars(namespace).items():
            # If the attribute is another SimpleNamespace, recursively initialize it as an instance of this class
            if isinstance(value, SimpleNamespace):
                value = RecursiveNamespaceToClass(value)
            # Set the attribute on the instance
            setattr(self, key, value)

    def __repr__(self):
        return f"{self.__class__.__name__}({vars(self)})"
    
    
z = zelda.Sensor('BALDR_UT_J3')



################## 
# configure our zwfs 
wvl0 = 1.25e-6
grid_dict = {
    "D":8, # diameter of beam (m)
    "N" : z.pupil_diameter, # number of pixels across pupil diameter
    "padding_factor" : z.pupil_dim // z.pupil_diameter, # how many pupil diameters fit into grid x axis
    # TOTAL NUMBER OF PIXELS = padding_factor * N 
    }

optics_dict = {
    "wvl0" :wvl0 , # central wavelength (m) 
    "F_number": z.mask_Fratio   , # F number on phasemask
    "mask_diam": z.mask_diameter / (1.22 * z.mask_Fratio * wvl0 ), # diameter of phaseshifting region in diffraction limit units (physical unit is mask_diam * 1.22 * F_number * lambda)
    "theta": z.mask_phase_shift( wvl0 )  # phaseshift of phasemask 
}

dm_dict = {
    "dm_model":"BMC-multi-3.5",
    "actuator_coupling_factor":0.7,# std of in actuator spacing of gaussian IF applied to each actuator. (e.g actuator_coupling_factor = 1 implies std of poke is 1 actuator across.)
    "dm_pitch":1,
    "dm_aoi":0, # angle of incidence of light on DM 
    "opd_per_cmd" : 3e-6, # peak opd applied at center of actuator per command unit (normalized between 0-1) 
    "flat_rmse" : 20e-9 # std (m) of flatness across Flat DM  
    }

grid_ns = SimpleNamespace(**grid_dict)
optics_ns = SimpleNamespace(**optics_dict)
dm_ns = SimpleNamespace(**dm_dict)

zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)

opd_input = zwfs_ns.grid.pupil_mask *  np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

opd_internal = 0 * zwfs_ns.grid.pupil_mask * np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

amp_input = 1e4 * zwfs_ns.grid.pupil_mask



# ++++++++++++++++++++++++++++++++++
# Zernike mask parameters
# ++++++++++++++++++++++++++++++++++

mask_substrate = 'fused_silica'
n_substrate = ztools.refractive_index(zwfs_ns.optics.wvl0 , mask_substrate)
# physical diameter and depth, in m
d_m = zwfs_ns.optics.mask_diam * ( 1.22 * zwfs_ns.optics.wvl0 * zwfs_ns.optics.F_number )
z_m = zwfs_ns.optics.theta * (zwfs_ns.optics.wvl0 / (2 * np.pi * (n_substrate - 1) ) )


I0 = ztools.propagate_opd_map(opd_map= 0 * opd_input, mask_diameter = d_m, mask_depth = z_m, mask_substrate = mask_substrate, mask_Fratio=zwfs_ns.optics.F_number,
                      pupil_diameter=zwfs_ns.grid.N, pupil = amp_input**0.5 , wave = zwfs_ns.optics.wvl0)


I = ztools.propagate_opd_map(opd_map= opd_input, mask_diameter = d_m, mask_depth = z_m, mask_substrate = mask_substrate, mask_Fratio=zwfs_ns.optics.F_number,
                      pupil_diameter=zwfs_ns.grid.N, pupil = amp_input**0.5 , wave = zwfs_ns.optics.wvl0)





#### TESTING MFT 
phi = zwfs_ns.grid.pupil_mask * 0* np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

pupil = 1e4 * zwfs_ns.grid.pupil_mask

theta = zwfs_ns.optics.theta 
psi_A = pupil * np.exp( 1j * ( phi ) )

# R_mask: mask radius in lam0/D unit
R_mask = 0.5 * zwfs_ns.optics.mask_diam

array_dim = phi.shape[0]
pupil_radius = zwfs_ns.grid.N // 2


# mask sampling in the focal plane
D_mask_pixels = 300

# ++++++++++++++++++++++++++++++++++
# Numerical simulation part
# ++++++++++++++++++++++++++++++++++

# --------------------------------
# plane A (Entrance pupil plane)

# definition of m1 parameter for the Matrix Fourier Transform (MFT)
# here equal to the mask size
mmm = 10

m1 = mmm * 2 * R_mask * (array_dim / (2. * pupil_radius))

ampl_PA = pupil * np.exp(1j * 2. * np.pi * phi / zwfs_ns.optics.wvl0)

ampl_PB = mft.mft(ampl_PA, array_dim, D_mask_pixels, m1)

phasemask_mask = aperture.disc(D_mask_pixels, D_mask_pixels//mmm, diameter=True, cpix=True, strict=False)
             
plt.figure(1)
plt.imshow( abs(ampl_PB) , extent=[-1, 1, -1, 1], origin='lower' )
X, Y = np.meshgrid( np.linspace(-1, 1, ampl_PB.shape[0]), np.linspace(-1, 1, ampl_PB.shape[0]) )
plt.colorbar()
plt.contour(X, Y, phasemask_mask, levels=10, colors='white', label='phasemask')
plt.show()
b = mft.imft(  phasemask_mask * ampl_PB , D_mask_pixels, array_dim, m1)

plt.figure(1)
plt.imshow( abs( b ), extent=[-1, 1, -1, 1], origin='lower' )
X, Y = np.meshgrid( np.linspace(-1, 1, b.shape[0]), np.linspace(-1, 1, b.shape[0]) )
plt.colorbar()
plt.contour(X, Y, abs( ampl_PA), levels=10, colors='white', label='phasemask')
plt.show()




psi_R = abs( b ) * np.sqrt((np.cos(theta)-1)**2 + np.sin(theta)**2)
mu = np.angle((np.exp(1J*theta)-1) ) # np.arctan( np.sin(theta)/(np.cos(theta)-1) ) #
beta = np.angle( b )
# out formula ----------
#if measured_pupil!=None:
#    P = measured_pupil / np.mean( P[P > np.mean(P)] ) # normalize by average value in Pupil

Ic = abs(psi_A)**2 + abs(psi_R)**2 + 2 * abs(psi_A) * abs(psi_R) * np.cos( phi - mu - beta )  #+ beta)

IcpyZ = z.propagate_opd_map(phi*1e9, mask_diameter, mask_depth, mask_substrate, mask_Fratio,
                      pupil_diameter, pupil, wave)


np.sum( abs(ampl_PA)**2 ) / np.sum( Ic )











Ime = bldr.get_frame( opd_input,  amp_input ,  opd_internal,  zwfs_ns )

fig,ax = plt.subplots( 3, 1, figsize=(10, 10) )
ax[0].imshow( I )
ax[1].imshow( Ime )
ax[2].imshow( Ime - I )
plt.show()



reference_wave, expi = create_reference_wave_beyond_pupil(mask_diameter=d_m, mask_depth=z_m, mask_substrate= mask_substrate,
                                                              mask_Fratio=zwfs_ns.optics.F_number, pupil_diameter=zwfs_ns.grid.N, pupil=zwfs_ns.grid.pupil_mask, \
                                                                  wave = zwfs_ns.optics.wvl0,
                                                              clear=clear, sign_mask=sign_mask, cpix=cpix)



plt.imshow( I - I0 ) ; plt.show()



propagate_opd_map(opd_map, mask_diameter, mask_depth, mask_substrate, mask_Fratio,
                      pupil_diameter, pupil, wave):

# substrate refractive index
n_substrate = ztools.refractive_index(zwfs_ns.optics.wvl0 , mask_substrate)

# R_mask: mask radius in lam0/D unit
R_mask = 0.5 * d_m / (zwfs_ns.optics.wvl0 * zwfs_ns.optics.F_number)

# ++++++++++++++++++++++++++++++++++
# Dimensions
# ++++++++++++++++++++++++++++++++++

# array and pupil
array_dim = zwfs_ns.grid.pupil_mask.shape[-1]
pupil_radius = zwfs_ns.grid.D // 2

# mask sampling in the focal plane
D_mask_pixels = 300








I = ztools.propagate_opd_map(opd_map= opd_input, mask_diameter, mask_depth, mask_substrate, mask_Fratio,
                      pupil_diameter=zwfs_ns.grid.D, pupil = zwfs_ns.grid.pupil_mask, wave = zwfs_ns.optics.wvl0)



wave = 1.642e-6

z = zelda.Sensor('SPHERE-IRDIS')

#################################

# Aberration: astigmatism
basis = np.nan_to_num(zernike.zernike_basis(nterms=5, npix=z.pupil_diameter))*1e-9
aberration = 20*basis[4]
gaussian_std = 1

# Gaussian apodized entrance pupil
x_vals = np.arange(z.pupil_diameter)
xx, yy = np.meshgrid(x_vals, x_vals)
cx, cy = x_vals[z.pupil_diameter//2], x_vals[z.pupil_diameter//2]
r = np.sqrt((xx-cx)**2+(yy-cy)**2)
r = r/r[0, z.pupil_diameter//2]

apodizer = np.exp(-(r**2)/2/gaussian_std**2)

#%%#############################
# Simulation of the wavefront. #
################################

# Clear pupil image
clear_pupil = abs(apodizer)**2

zelda_pupil0 = ztools.propagate_opd_map(0*aberration, z.mask_diameter, z.mask_depth, z.mask_substrate,
                                        z.mask_Fratio, z.pupil_diameter, apodizer, wave)

zelda_pupil = ztools.propagate_opd_map(aberration, z.mask_diameter, z.mask_depth, z.mask_substrate,
                                        z.mask_Fratio, z.pupil_diameter, apodizer, wave)



pupil_roi = aperture.disc(z.pupil_diameter, z.pupil_diameter, diameter=True, cpix=False)
z_opd_advanced = z.analyze(clear_pupil, zelda_pupil, wave,
                           use_arbitrary_amplitude=True,
                           refwave_from_clear=True,
                           cpix=False, pupil_roi=pupil_roi)

#%% plot

fig = plt.figure(0, figsize=(24, 4))
plt.clf()

gs = gridspec.GridSpec(ncols=7, nrows=1, figure=fig, width_ratios=[.1,1,1,1,1,1,.1])

ax = fig.add_subplot(gs[0,1])
mappable = ax.imshow(clear_pupil, aspect='equal', vmin=0, vmax=1)
ax.set_title('Clear pupil')

ax = fig.add_subplot(gs[0,0])
cbar1 = fig.colorbar(mappable=mappable, cax=ax)
cbar1.set_label('Normalized intensity')

ax = fig.add_subplot(gs[0,2])
ax.imshow(zelda_pupil, aspect='equal', vmin=0, vmax=1)
ax.set_title('ZELDA pupil')

ax = fig.add_subplot(gs[0,3])
ax.imshow(aberration*1e9, aspect='equal', vmin=-40, vmax=40, cmap='magma')
ax.set_title('Introduced aberration (nm)')

ax = fig.add_subplot(gs[0,4])
cax = ax.imshow(z_opd_standard, aspect='equal', vmin=-40, vmax=40, cmap='magma')
ax.set_title('Reconstructed aberration - standard')

ax = fig.add_subplot(gs[0,5])
cax = ax.imshow(z_opd_advanced, aspect='equal', vmin=-40, vmax=40, cmap='magma')
ax.set_title('Reconstructed aberration - advanced')

ax = fig.add_subplot(gs[0,6])
cbar = fig.colorbar(mappable=cax, cax=ax)
cbar.set_label('OPD [nm]')

plt.tight_layout()
plt.show()


#!/usr/bin/env python3
import streamlit as st
from streamlit.web import cli as stcli
from streamlit.runtime.scriptrunner import get_script_run_ctx

import numpy as np
import poppy
import matplotlib.pyplot as plt
from poppy.poppy_core import PlaneType
from poppy.zernike import zernike
import astropy.units as u
import os 
import sys
st.write("Starting app...")

# Thin lens functions and system
class lens:
    def __init__(self, x, f):
        self.f = f
        self.x = x
        
    def get_image(self, _object):
        do = (self.x - _object.x)
        if do != self.f:
            di = do * self.f / (do - self.f)
        else:
            di = np.inf
        
        M = -di/do
        i = obj(x=self.x + di, h=M * _object.h)
        return i

class obj:
    def __init__(self, x, h):
        self.x = x
        self.h = h

def propagate_object(lens_list, _object):
    object_list = [_object]
    for l in lens_list:
        next_object = l.get_image(object_list[-1])
        object_list.append(next_object)
    return object_list

# Function to bin the intensity image
def bin_image(image, bin_size):
    shape = (image.shape[0] // bin_size, bin_size,
             image.shape[1] // bin_size, bin_size)
    binned_image = image.reshape(shape).mean(-1).mean(1)
    return binned_image

# Define a custom optical element to apply Zernike aberration
class ZernikeAberration(poppy.AnalyticOpticalElement):
    def __init__(self, phase_aberration, name="Zernike Aberration"):
        super().__init__(name=name, planetype=poppy.poppy_core.PlaneType.pupil)
        self.phase_aberration = phase_aberration

    def get_opd(self, wave):
        return self.phase_aberration / (2 * np.pi / wave.wavelength.to(u.m).value)

# Phase Mask element
class PhaseMask(poppy.AnalyticOpticalElement):
    def __init__(self, D_phasemask, theta, T_on, T_off, name='Phase Mask'):
        super().__init__(name=name, planetype=PlaneType.pupil)
        self.D_phasemask = D_phasemask
        self.theta = theta
        self.T_on = T_on
        self.T_off = T_off

    def get_transmission(self, wave):
        y, x = wave.coordinates()  # 2D coordinate grid
        r = np.sqrt(x**2 + y**2)
        phase_mask = np.ones_like(r, dtype=complex)
        inside_mask = r <= (self.D_phasemask / 2)
        outside_mask = r > (self.D_phasemask / 2)
        phase_mask[inside_mask] = self.T_on * np.exp(1j * np.deg2rad(self.theta))
        phase_mask[outside_mask] = self.T_off
        return phase_mask

# Sidebar for user inputs
st.sidebar.title("Zernike Wavefront Sensor Inputs")

st.write("Instructions: Adjust alignment parameters in the side panel and press 'Update'. No image will be shown until you press update") # for logging

st.write("Reading in user variables.. Don't forget to press update on the side panel!") # for logging

# Initialize session_state
if "submitted" not in st.session_state:
    st.session_state.submitted = False
    
# Submit button
# Update button
if st.sidebar.button("Update"):
    st.write("changing session state..")
    st.session_state.submitted = True

# Input field inputs
wavelength = st.sidebar.number_input("Wavelength (um)", value=1.25)
zernike_mode = st.sidebar.number_input("Zernike Mode", value=None, step=1)
zernike_coefficient = st.sidebar.number_input("Zernike Coefficient", value=0.08)

# Phasemask inputs
phasemask_diameter = st.sidebar.number_input("Phasemask Diameter (lambda/D)", value=1.06)
theta = st.sidebar.number_input("Phase Mask Shift (degrees)", value=90)
T_on = st.sidebar.number_input("Phasemask On-Axis Transmission", value=1.0)
T_off = st.sidebar.number_input("Phasemask Off-Axis Transmission", value=1.0)

# Offset inputs
phasemask_offset = 1e-6 * st.sidebar.number_input("Phasemask Offset (um)", value=0.0)
lens1_offset = 1e-6 * st.sidebar.number_input("Collimating Lens Offset (um)", value=0.0)
lens2_offset = 1e-6 * st.sidebar.number_input("Pupil Imaging Lens Offset (um)", value=0.0)
coldstop_offset = 1e-6 * st.sidebar.number_input("Cold Stop Offset (um)", value=0.0)
detector_offset = 1e-6 * st.sidebar.number_input("Detector Offset (um)", value=0.0)

# offset_input_dict = {
#     "Phasemask Offset":phasemask_offset,
#     "Collimating Lens Offset":lens1_offset,
#     "Pupil Imaging Lens Offset":lens2_offset,
#     "Cold Stop Offset":coldstop_offset,
#     "Detector Offset":detector_offset
# }

# Include elements toggle switches
include_phase_mask = st.sidebar.checkbox("Include Phase Mask", value=True)
include_cold_stop = st.sidebar.checkbox("Include Cold Stop", value=True)



# Only generate plot if Update button is clicked
if st.session_state.submitted:
    st.write("updating to plot output intensity (after submit)..")
    # Hardcoded values
    npix = 256
    D = 12e-3  # Pupil diameter in meters
    f_oap = 254e-3  # OAP mirror focal length in meters
    f_lens1 = 30e-3  # Focal length of lens 1
    f_lens2 = 200e-3  # Focal length of lens 2
    cold_stop_diameter = 2.15e-3  # 2.15 mm
    distance_l1_to_l2 = 1228.570e-3

    # Thin lens system calculations for element positions
    x0 = -1  # Object position
    l1 = lens(x=0, f=f_oap)
    l2 = lens(x=f_oap + f_lens1, f=f_lens1)
    l3 = lens(x=f_oap + f_lens1 + distance_l1_to_l2, f=f_lens2)
    sys = [l1, l2, l3]

    oPup = obj(x=x0, h=D/2)
    pup_ims = propagate_object(sys, oPup)
    z_detector = pup_ims[-1].x

    # Absolute positions of the elements
    z_oap = 0
    z_fpm = f_oap
    z_lens1 = f_oap + f_lens1
    z_lens2 = f_oap + f_lens1 + distance_l1_to_l2
    z_coldstop = f_oap + f_lens1 + distance_l1_to_l2 + f_lens2
    abs_pos = np.array([0, z_fpm, z_lens1, z_lens2, z_coldstop, z_detector])
    
    # Calculated variables
    wavelength = wavelength * 1e-6 # convert um to m 
    D_phasemask = phasemask_diameter * 1.22 * wavelength * f_oap / D

    # Create optical elements
    oap_mirror = poppy.QuadraticLens(f_lens=f_oap * u.m, name='OAP Mirror')
    phase_mask = PhaseMask(D_phasemask=D_phasemask, theta=theta, T_on=T_on, T_off=T_off)
    lens1 = poppy.QuadraticLens(f_lens=f_lens1 * u.m, name='Lens 1')
    lens2 = poppy.QuadraticLens(f_lens=f_lens2 * u.m, name='Lens 2')
    cold_stop = poppy.CircularAperture(radius=cold_stop_diameter / 2 * u.m, name='Cold Stop')



    element_list = [oap_mirror, phase_mask, lens1, lens2, cold_stop] # we need to also explicitly define the full element_list for later
    # usr_element_list is primarily for creating usr_include_list which we also need later
    if (include_cold_stop) and (not include_phase_mask):
        usr_element_list = [oap_mirror, lens1, lens2, cold_stop]
    elif (not include_cold_stop) and (include_phase_mask):
        usr_element_list = [oap_mirror, phase_mask, lens1, lens2]
    elif (not include_cold_stop) and (not include_phase_mask):
        usr_element_list = [oap_mirror, lens1, lens2]
    else:
        usr_element_list = [oap_mirror, phase_mask, lens1, lens2, cold_stop]

    usr_include_list = [e.name for e in usr_element_list] 
    offset_list = [phasemask_offset, lens1_offset,lens2_offset, coldstop_offset, detector_offset] # MUST BE SAME ORDER AS element_list (without user removal!)
    

    abs_pos_with_offset = abs_pos.copy()
    for i, offset in enumerate( offset_list ):
        abs_pos_with_offset[i+1] += offset # i+1 since offset list does not include an offset for the first element (oap)
        
    rel_pos_with_offsets = np.diff( abs_pos_with_offset )

    wf = poppy.FresnelWavefront(beam_radius=D/2 *u.m, wavelength=wavelength *u.m, npix=npix, oversample=4)
    wf *= poppy.CircularAperture(radius=D/2) 

    # Apply Zernike aberration
    if zernike_mode is not None:
        y, x = wf.coordinates()
        r = np.sqrt(x**2 + y**2) / D * 2
        theta_vals = np.arctan2(y, x)
        n, m = poppy.zernike.noll_indices(zernike_mode)
        zernike_phase = zernike(n=n, m=m, rho=r, theta=theta_vals, outside=0)
        aberration_phase = 2 * np.pi * zernike_coefficient * zernike_phase
        zernike_aberration_element = ZernikeAberration(aberration_phase)
        wf *= zernike_aberration_element


    wf.propagate_fresnel( x0 * u.m ) 
    for element, rel_dist in zip( element_list ,   rel_pos_with_offsets ):
        # since we deal with relative positions we MUST go through all the elements (hence why we kept element_list above despite usr input)
        # if the user did not want to include one we still propagate the relative distance, we just 
        # don't aspply the elements transformation.
        if element.name in usr_include_list: # sometimes we might want to exclude an element (like the phasemask, or cold stop to analyse its effect)
            print( element.name )
            wf *= element
            
        wf.propagate_fresnel( rel_dist * u.m )
        
    # Get intensity at the detector
    intensity = bin_image(abs(wf.amplitude)**2, bin_size=16)

    st.write("Plotting..")
    # Display intensity image
    plt.figure()
    plt.imshow(intensity, cmap='inferno')
    plt.colorbar(label='Intensity')
    plt.title("Intensity at the Detector")
    st.pyplot(plt)



def main():
    # Check if the script is already running in Streamlit
    if get_script_run_ctx() is not None:
        # Your Streamlit app code goes here
        import streamlit as st
        st.title("Baldr Fresnel App")
        st.write("Welcome to the Baldr Fresnel App!")
    else:
        # Launch the app using streamlit run
        app_path = os.path.abspath(__file__)
        sys.argv = ["streamlit", "run", app_path]
        stcli.main()

if __name__ == "__main__":
    main()


import numpy as np
import matplotlib.pyplot as plt
import os 
import sys
import pandas as pd
from scipy.optimize import curve_fit


# Define Cauchy's equation
def cauchy_eqn(wavelength, A, B, C):
    return A + B / wavelength**2 + C / wavelength**4


def fit_cauchy_eqn_to_data(df, savefig = None):
    # Example dataframe (replace with your actual df)
    # df = pd.read_csv('path_to_your_data.csv')  # If your data is stored in a CSV
    # Assuming your data is already in a dataframe 'df'
    wavelengths = df['Wavelength(um)'].values  # Extracting wavelength values
    n_measured = df['n'].values  # Extracting refractive index values

    # Perform the curve fitting
    popt, pcov = curve_fit(cauchy_eqn, wavelengths, n_measured)

    # Extract the fitted coefficients
    A_fit, B_fit, C_fit = popt

    # Generate the fitted curve
    n_fitted = cauchy_eqn(wavelengths, A_fit, B_fit, C_fit)

    # Plot the measured data vs the fitted curve
    plt.plot(wavelengths, n_measured, 'b-', label='Measured Data')
    plt.plot(wavelengths, n_fitted, 'r--', label=f'Fitted Cauchy Eqn\nA={A_fit:.4f}, B={B_fit:.4e}, C={C_fit:.4e}')
    plt.xlabel('Wavelength [um]')
    plt.ylabel('Refractive Index n')
    plt.legend()
    if savefig is not None:
        plt.savefig(savefig, bbox_inches='tight', dpi=200)  
    plt.show()


df = pd.read_csv('data/Exposed_Ma-N_1405_optical_constants.txt', sep='\s+', header=1)
df['Wavelength(um)'] =  df['Wavelength(nm)'] * 1e-3
df_900nm_cut = df[df['Wavelength(nm)'] > 900]
fit_cauchy_eqn_to_data( df, savefig = '/home/benja/Documents/BALDR/figures/Exposed_Ma-N_1405_cauchy_fit.png' )

fit_cauchy_eqn_to_data( df_900nm_cut, savefig = '/home/benja/Documents/BALDR/figures/Exposed_Ma-N_1405_cauchy_fit_wave900-1600.png' )


"""
forked from aotools github https://github.com/AOtools/aotools/blob/main/aotools/turbulence/infinitephasescreen.py. 
Removed numba dependency due to conflicts.

Infinite Phase Screens
----------------------

An implementation of the "infinite phase screen", as deduced by Francois Assemat and Richard W. Wilson, 2006.
"""

from scipy import linalg
import numpy
from numpy import fft
from scipy.special import gamma, kv
#import numba



def ft_sh_phase_screen(r0, N, delta, L0, l0, FFT=None, seed=None):

    """
    Creates a random phase screen with Von Karmen statistics with added
    sub-harmonics to augment tip-tilt modes.
    (Schmidt 2010)

    .. note::
        The phase screen is returned as a 2d array, with each element representing the phase 
        change in **radians**. This means that to obtain the physical phase distortion in nanometres, 
        it must be multiplied by (wavelength / (2*pi)), (where `wavellength` here is the same wavelength
        in which r0 is given in the function arguments)

    Args:
        r0 (float): r0 parameter of scrn in metres
        N (int): Size of phase scrn in pxls
        delta (float): size in Metres of each pxl
        L0 (float): Size of outer-scale in metres
        l0 (float): inner scale in metres
        seed (int, optional): seed for random number generator. If provided, 
            allows for deterministic screens  

    Returns:
        ndarray: numpy array representing phase screen in radians
    """
    R = numpy.random.default_rng(seed)

    D = N * delta
    # high-frequency screen from FFT method
    phs_hi = ft_phase_screen(r0, N, delta, L0, l0, FFT, seed=seed)

    # spatial grid [m]
    coords = numpy.arange(-N/2,N/2)*delta
    x, y = numpy.meshgrid(coords,coords)

    # initialize low-freq screen
    phs_lo = numpy.zeros(phs_hi.shape)

    # loop over frequency grids with spacing 1/(3^p*L)
    for p in range(1,4):
        # setup the PSD
        del_f = 1 / (3**p*D) #frequency grid spacing [1/m]
        fx = numpy.arange(-1,2) * del_f

        # frequency grid [1/m]
        fx, fy = numpy.meshgrid(fx,fx)
        f = numpy.sqrt(fx**2 +  fy**2) # polar grid

        fm = 5.92/l0/(2*numpy.pi) # inner scale frequency [1/m]
        f0 = 1./L0

        # outer scale frequency [1/m]
        # modified von Karman atmospheric phase PSD
        PSD_phi = (0.023*r0**(-5./3)
                    * numpy.exp(-1*(f/fm)**2) / ((f**2 + f0**2)**(11./6)) )
        PSD_phi[1,1] = 0

        # random draws of Fourier coefficients
        cn = ( (R.normal(size=(3,3))
            + 1j*R.normal(size=(3,3)) )
                        * numpy.sqrt(PSD_phi)*del_f )
        SH = numpy.zeros((N,N),dtype="complex")
        # loop over frequencies on this grid
        for i in range(0, 3):
            for j in range(0, 3):

                SH += cn[i,j] * numpy.exp(1j*2*numpy.pi*(fx[i,j]*x+fy[i,j]*y))

        phs_lo = phs_lo + SH
        # accumulate subharmonics

    phs_lo = phs_lo.real - phs_lo.real.mean()

    phs = phs_lo+phs_hi

    return phs


def ft_phase_screen(r0, N, delta, L0, l0, FFT=None, seed=None):
    """
    Creates a random phase screen with Von Karmen statistics.
    (Schmidt 2010)
    
    Parameters:
        r0 (float): r0 parameter of scrn in metres
        N (int): Size of phase scrn in pxls
        delta (float): size in Metres of each pxl
        L0 (float): Size of outer-scale in metres
        l0 (float): inner scale in metres
        seed (int, optional): seed for random number generator. If provided, 
            allows for deterministic screens  

    .. note::
        The phase screen is returned as a 2d array, with each element representing the phase 
        change in **radians**. This means that to obtain the physical phase distortion in nanometres, 
        it must be multiplied by (wavelength / (2*pi)), (where `wavellength` here is the same wavelength
        in which r0 is given in the function arguments)

    Returns:
        ndarray: numpy array representing phase screen in radians
    """
    delta = float(delta)
    r0 = float(r0)
    L0 = float(L0)
    l0 = float(l0)

    R = numpy.random.default_rng(seed)

    del_f = 1./(N*delta)

    fx = numpy.arange(-N/2., N/2.) * del_f

    (fx, fy) = numpy.meshgrid(fx,fx)
    f = numpy.sqrt(fx**2. + fy**2.)

    fm = 5.92/l0/(2*numpy.pi)
    f0 = 1./L0

    PSD_phi = (0.023*r0**(-5./3.) * numpy.exp(-1*((f/fm)**2)) / (((f**2) + (f0**2))**(11./6)))

    PSD_phi[int(N/2), int(N/2)] = 0

    cn = ((R.normal(size=(N, N))+1j * R.normal(size=(N, N))) * numpy.sqrt(PSD_phi)*del_f)

    phs = ift2(cn, 1, FFT).real

    return phs


def ift2(G, delta_f, FFT=None):
    """
    Wrapper for inverse fourier transform

    Parameters:
        G: data to transform
        delta_f: pixel seperation
        FFT (FFT object, optional): An accelerated FFT object
    """

    N = G.shape[0]

    if FFT:
        g = numpy.fft.fftshift(FFT(numpy.fft.fftshift(G))) * (N * delta_f) ** 2
    else:
        g = fft.ifftshift(fft.ifft2(fft.fftshift(G))) * (N * delta_f) ** 2

    return g

def phase_covariance(r, r0, L0):
    """
    Calculate the phase covariance between two points seperated by `r`, 
    in turbulence with a given `r0 and `L0`.
    Uses equation 5 from Assemat and Wilson, 2006.

    Parameters:
        r (float, ndarray): Seperation between points in metres (can be ndarray)
        r0 (float): Fried parameter of turbulence in metres
        L0 (float): Outer scale of turbulence in metres
    """
    # Make sure everything is a float to avoid nasty surprises in division!
    r = numpy.float32(r)
    r0 = float(r0)
    L0 = float(L0)

    # Get rid of any zeros
    r += 1e-40

    A = (L0 / r0) ** (5. / 3)

    B1 = (2 ** (-5. / 6)) * gamma(11. / 6) / (numpy.pi ** (8. / 3))
    B2 = ((24. / 5) * gamma(6. / 5)) ** (5. / 6)

    C = (((2 * numpy.pi * r) / L0) ** (5. / 6)) * kv(5. / 6, (2 * numpy.pi * r) / L0)

    cov = A * B1 * B2 * C

    return cov


class PhaseScreen(object):
    """
    A "Phase Screen" for use in AO simulation.  Can be extruded infinitely.

    This represents the phase addition light experiences when passing through atmospheric 
    turbulence. Unlike other phase screen generation techniques that translate a large static 
    screen, this method keeps a small section of phase, and extends it as necessary for as many
    steps as required. This can significantly reduce memory consumption at the expense of more
    processing power required.

    The technique is described in a paper by Assemat and Wilson, 2006 and expanded upon by Fried, 2008.
    It essentially assumes that there are two matrices, "A" and "B",
    that can be used to extend an existing phase screen.
    A single row or column of new phase can be represented by 

        X = A.Z + B.b

    where X is the new phase vector, Z is some data from the existing screen,
    and b is a random vector with gaussian statistics.

    This object calculates the A and B matrices using an expression of the phase covariance when it
    is initialised. Calculating A is straightforward through the relationship:

        A =  Cov_xz . (Cov_zz)^(-1).

    B is less trivial.

        BB^t = Cov_xx - A.Cov_zx

    (where B^t is the transpose of B) is a symmetric matrix, hence B can be expressed as 

        B = UL, 

    where U and L are obtained from the svd for BB^t

        U, w, U^t = svd(BB^t)

    L is a diagonal matrix where the diagonal elements are w^(1/2).    

    On initialisation an initial phase screen is calculated using an FFT based method.
    When 'add_row' is called, a new vector of phase is added to the phase screen.

    Existing points to use are defined by a "stencil", than is set to 0 for points not to use
    and 1 for points to use. This makes this a generalised base class that can be used by 
    other infinite phase screen creation schemes, such as for Von Karmon turbulence or 
    Kolmogorov turbulence.

    .. note::
        The phase screen is returned on each iteration as a 2d array, with each element representing the phase 
        change in **radians**. This means that to obtain the physical phase distortion in nanometres, 
        it must be multiplied by (wavelength / (2*pi)), (where `wavellength` here is the same wavelength
        in which r0 is given in the function arguments)
    """
    def set_X_coords(self):
        """
        Sets the coords of X, the new phase vector.
        """
        self.X_coords = numpy.zeros((self.nx_size, 2))
        self.X_coords[:, 0] = -1
        self.X_coords[:, 1] = numpy.arange(self.nx_size)
        self.X_positions = self.X_coords * self.pixel_scale

    def set_stencil_coords(self):
        """
        Sets the Z coordinates, sections of the phase screen that will be used to create new phase
        """
        self.stencil = numpy.zeros((self.stencil_length, self.nx_size))

        max_n = 1
        while True:
            if 2 ** (max_n - 1) + 1 >= self.nx_size:
                max_n -= 1
                break
            max_n += 1

        for n in range(0, max_n + 1):
            col = int((2 ** (n - 1)) + 1)
            n_points = (2 ** (max_n - n)) + 1

            coords = numpy.round(numpy.linspace(0, self.nx_size - 1, n_points)).astype('int32')
            self.stencil[col - 1][coords] = 1

        # Now fill in tail of stencil
        for n in range(1, self.stencil_length_factor + 1):
            col = n * self.nx_size - 1
            self.stencil[col, self.nx_size // 2] = 1

        self.stencil_coords = numpy.array(numpy.where(self.stencil == 1)).T
        self.stencil_positions = self.stencil_coords * self.pixel_scale

        self.n_stencils = len(self.stencil_coords)

    def calc_seperations(self):
        """
        Calculates the seperations between the phase points in the stencil and the new phase vector
        """
        positions = numpy.append(self.stencil_positions, self.X_positions, axis=0)
        self.seperations = numpy.zeros((len(positions), len(positions)))

        #if numba:
        #    calc_seperations_fast(positions, self.seperations)
        if 1: #else:
            for i, (x1, y1) in enumerate(positions):
                for j, (x2, y2) in enumerate(positions):
                    delta_x = x2 - x1
                    delta_y = y2 - y1

                    delta_r = numpy.sqrt(delta_x ** 2 + delta_y ** 2)

                    self.seperations[i, j] = delta_r



    def make_covmats(self):
        """
        Makes the covariance matrices required for adding new phase
        """
        self.cov_mat = phase_covariance(self.seperations, self.r0, self.L0)

        self.cov_mat_zz = self.cov_mat[:self.n_stencils, :self.n_stencils]
        self.cov_mat_xx = self.cov_mat[self.n_stencils:, self.n_stencils:]
        self.cov_mat_zx = self.cov_mat[:self.n_stencils, self.n_stencils:]
        self.cov_mat_xz = self.cov_mat[self.n_stencils:, :self.n_stencils]

    def makeAMatrix(self):
        """
        Calculates the "A" matrix, that uses the existing data to find a new 
        component of the new phase vector.
        """
        # Cholsky solve can fail - if so do brute force inversion
        try:
            cf = linalg.cho_factor(self.cov_mat_zz)
            inv_cov_zz = linalg.cho_solve(cf, numpy.identity(self.cov_mat_zz.shape[0]))
        except linalg.LinAlgError:
            # print("Cholesky solve failed. Performing SVD inversion...")
            # inv_cov_zz = numpy.linalg.pinv(self.cov_mat_zz)
            raise linalg.LinAlgError("Could not invert Covariance Matrix to for A and B Matrices. Try with a larger pixel scale or smaller L0")

        self.A_mat = self.cov_mat_xz.dot(inv_cov_zz)

    def makeBMatrix(self):
        """
        Calculates the "B" matrix, that turns a random vector into a component of the new phase.
        """
        # Can make initial BBt matrix first
        BBt = self.cov_mat_xx - self.A_mat.dot(self.cov_mat_zx)

        # Then do SVD to get B matrix
        u, W, ut = numpy.linalg.svd(BBt)

        L_mat = numpy.zeros((self.nx_size, self.nx_size))
        numpy.fill_diagonal(L_mat, numpy.sqrt(W))

        # Now use sqrt(eigenvalues) to get B matrix
        self.B_mat = u.dot(L_mat)

    def make_initial_screen(self):
        """
        Makes the initial screen usign FFT method that can be extended 
        """

        # phase screen will make it *really* random if no seed at all given.
        # If a seed is here, screen must be repeatable wiht same seed
        self._R = numpy.random.default_rng(self.random_seed)

        self._scrn = ft_phase_screen(
            self.r0, self.stencil_length, self.pixel_scale, self.L0, 1e-10, seed=self._R
        )

        self._scrn = self._scrn[:, :self.nx_size]

    def get_new_row(self):
        random_data = self._R.normal(0, 1, size=self.nx_size)

        stencil_data = self._scrn[(self.stencil_coords[:, 0], self.stencil_coords[:, 1])]
        new_row = self.A_mat.dot(stencil_data) + self.B_mat.dot(random_data)

        new_row.shape = (1, self.nx_size)
        return new_row

    def add_row(self):
        """
        Adds a new row to the phase screen and removes old ones.
        """

        new_row = self.get_new_row()

        self._scrn = numpy.append(new_row, self._scrn, axis=0)[:self.stencil_length, :self.nx_size]

        return self.scrn

    @property
    def scrn(self):
        """
        The current phase map held in the PhaseScreen object in radians.
        """
        return self._scrn[:self.requested_nx_size, :self.requested_nx_size]


class PhaseScreenVonKarman(PhaseScreen):
    """
    A "Phase Screen" for use in AO simulation with Von Karmon statistics.

    This represents the phase addition light experiences when passing through atmospheric
    turbulence. Unlike other phase screen generation techniques that translate a large static
    screen, this method keeps a small section of phase, and extends it as necessary for as many
    steps as required. This can significantly reduce memory consumption at the expense of more
    processing power required.

    The technique is described in a paper by Assemat and Wilson, 2006. It essentially assumes that
    there are two matrices, "A" and "B", that can be used to extend an existing phase screen.
    A single row or column of new phase can be represented by

        X = A.Z + B.b

    where X is the new phase vector, Z is some number of columns of the existing screen,
    and b is a random vector with gaussian statistics.

    This object calculates the A and B matrices using an expression of the phase covariance when it
    is initialised. Calculating A is straightforward through the relationship:

        A =  Cov_xz . (Cov_zz)^(-1).

    B is less trivial.

        BB^t = Cov_xx - A.Cov_zx

    (where B^t is the transpose of B) is a symmetric matrix, hence B can be expressed as

        B = UL,

    where U and L are obtained from the svd for BB^t

        U, w, U^t = svd(BB^t)

    L is a diagonal matrix where the diagonal elements are w^(1/2).

    On initialisation an initial phase screen is calculated using an FFT based method.
    When ``add_row`` is called, a new vector of phase is added to the phase screen using `nCols`
    columns of previous phase. Assemat & Wilson claim that two columns are adequate for good
    atmospheric statistics. The phase in the screen data is always accessed as ``<phasescreen>.scrn`` and is in radians.

        .. note::
        The phase screen is returned on each iteration as a 2d array, with each element representing the phase 
        change in **radians**. This means that to obtain the physical phase distortion in nanometres, 
        it must be multiplied by (wavelength / (2*pi)), (where `wavellength` here is the same wavelength
        in which r0 is given in the function arguments)

    Parameters:
        nx_size (int): Size of phase screen (NxN)
        pixel_scale(float): Size of each phase pixel in metres
        r0 (float): fried parameter (metres)
        L0 (float): Outer scale (metres)
        random_seed (int, optional): seed for the random number generator
        n_columns (int, optional): Number of columns to use to continue screen, default is 2
    """
    def __init__(self, nx_size, pixel_scale, r0, L0, random_seed=None, n_columns=2):

        self.n_columns = n_columns

        self.requested_nx_size = nx_size
        self.nx_size = nx_size
        self.pixel_scale = pixel_scale
        self.r0 = r0
        self.L0 = L0
        self.stencil_length_factor = 1
        self.stencil_length = self.nx_size

        self.random_seed = random_seed

        self.set_X_coords()
        self.set_stencil_coords()

        self.calc_seperations()
        self.make_covmats()

        self.makeAMatrix()
        self.makeBMatrix()
        self.make_initial_screen()


    def set_stencil_coords(self):
        self.stencil = numpy.zeros((self.stencil_length, self.nx_size))
        self.stencil[:self.n_columns] = 1

        self.stencil_coords = numpy.array(numpy.where(self.stencil==1)).T
        self.stencil_positions = self.stencil_coords * self.pixel_scale

        self.n_stencils = len(self.stencil_coords)


def find_allowed_size(nx_size):
    """
    Finds the next largest "allowed size" for the Fried Phase Screen method
    
    Parameters:
        nx_size (int): Requested size
    
    Returns:
        int: Next allowed size
    """
    n = 0
    while (2 ** n + 1) < nx_size:
        n += 1

    nx_size = 2 ** n + 1
    return nx_size


class PhaseScreenKolmogorov(PhaseScreen):
    """
    A "Phase Screen" for use in AO simulation using the Fried method for Kolmogorov turbulence.

    This represents the phase addition light experiences when passing through atmospheric 
    turbulence. Unlike other phase screen generation techniques that translate a large static 
    screen, this method keeps a small section of phase, and extends it as neccessary for as many 
    steps as required. This can significantly reduce memory consuption at the expense of more 
    processing power required.

    The technique is described in a paper by Assemat and Wilson, 2006 and expanded upon by Fried, 2008.
    It essentially assumes that there are two matrices, "A" and "B",
    that can be used to extend an existing phase screen.
    A single row or column of new phase can be represented by 

        X = A.Z + B.b

    where X is the new phase vector, Z is some data from the existing screen,
    and b is a random vector with gaussian statistics.

    This object calculates the A and B matrices using an expression of the phase covariance when it
    is initialised. Calculating A is straightforward through the relationship:

        A =  Cov_xz . (Cov_zz)^(-1).

    B is less trivial.

        BB^t = Cov_xx - A.Cov_zx

    (where B^t is the transpose of B) is a symmetric matrix, hence B can be expressed as 

        B = UL, 

    where U and L are obtained from the svd for BB^t

        U, w, U^t = svd(BB^t)

    L is a diagonal matrix where the diagonal elements are w^(1/2).    

    The Z data is taken from points in a "stencil" defined by Fried that samples the entire screen.
    An additional "reference point" is also considered, that is picked from a point separate from teh stencil 
    and applied on each iteration such that the new phase equation becomes:
    
    On initialisation an initial phase screen is calculated using an FFT based method.
    When ``add_row`` is called, a new vector of phase is added to the phase screen. The phase in the screen data
    is always accessed as ``<phasescreen>.scrn`` and is in radians.

    .. note::
        The phase screen is returned on each iteration as a 2d array, with each element representing the phase 
        change in **radians**. This means that to obtain the physical phase distortion in nanometres, 
        it must be multiplied by (wavelength / (2*pi)), (where `wavellength` here is the same wavelength
        in which r0 is given in the function arguments)

    Parameters:
        nx_size (int): Size of phase screen (NxN)
        pixel_scale(float): Size of each phase pixel in metres
        r0 (float): fried parameter (metres)
        L0 (float): Outer scale (metres)
        random_seed (int, optional): seed for the random number generator
        stencil_length_factor (int, optional): How much longer is the stencil than the desired phase? default is 4
    """
    def __init__(self, nx_size, pixel_scale, r0, L0, random_seed=None, stencil_length_factor=4):

        self.requested_nx_size = nx_size
        self.nx_size = find_allowed_size(nx_size)
        self.pixel_scale = pixel_scale
        self.r0 = r0
        self.L0 = L0
        self.stencil_length_factor = stencil_length_factor
        self.stencil_length = stencil_length_factor * self.nx_size
        self.random_seed = random_seed

        # Coordinate of Fried's "reference point" that stops the screen diverging
        self.reference_coord = (1, 1)

        self.set_X_coords()
        self.set_stencil_coords()

        self.calc_seperations()
        self.make_covmats()

        self.makeAMatrix()
        self.makeBMatrix()
        self.make_initial_screen()

    def get_new_row(self):
        random_data = self._R.normal(0, 1, size=self.nx_size)

        stencil_data = self._scrn[(self.stencil_coords[:, 0], self.stencil_coords[:, 1])]

        reference_value = self._scrn[self.reference_coord]

        new_row = self.A_mat.dot(stencil_data - reference_value) + self.B_mat.dot(random_data) + reference_value

        new_row.shape = (1, self.nx_size)
        return new_row


    def __repr__(self):
        return str(self.scrn)
    


#@numba.jit(nopython=True, parallel=True)
# def calc_seperations_fast(positions, seperations):

#     for i in numba.prange(len(positions)):
#         x1, y1 = positions[i]
#         for j in range(len(positions)):
#             x2, y2 = positions[j]
#             delta_x = x2 - x1
#             delta_y = y2 - y1

#             delta_r = numpy.sqrt(delta_x ** 2 + delta_y ** 2)

#             seperations[i, j] = delta_r


"""
Lets see what the optical gain looks like for a simple system rolling Kolmogorov phase screens at different Strehl ratios across the pupil.

explore the real and imaginary components of the optical gain both inside and outside the pupil.

"""
import numpy as np
import matplotlib.pyplot as plt
from types import SimpleNamespace
import importlib 
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid1 import make_axes_locatable
import os 
import sys
#import aotools

def add_project_root_to_sys_path(project_root_name="BaldrApp"):
    """
    Adds the project root directory to sys.path to allow importing from shared modules.
    
    Args:
        project_root_name (str): The name of the project root directory.
    """
    try:
        # Attempt to use __file__ to get the current script's directory
        current_dir = os.path.dirname(os.path.abspath(__file__))
    except NameError:
        # Fallback to current working directory (useful in interactive environments)
        current_dir = os.getcwd()

    # Traverse up the directory tree to find the project root
    project_root = current_dir
    while True:
        if os.path.basename(project_root) == project_root_name:
            break
        new_project_root = os.path.dirname(project_root)
        if new_project_root == project_root:
            # Reached the filesystem root without finding the project root
            project_root = None
            break
        project_root = new_project_root

    if project_root and project_root not in sys.path:
        sys.path.append(project_root)
        print(f"Added '{project_root}' to sys.path")
    elif project_root is None:
        print(f"Error: '{project_root_name}' directory not found in the directory hierarchy.")
    else:
        print(f"'{project_root}' is already in sys.path")

# Call the function to add the project root
add_project_root_to_sys_path()
from baldrapp.common import baldr_core as bldr
from baldrapp.common import DM_basis as gen_basis
from baldrapp.common import utilities as util
from baldrapp.common import DM_registration as DM_reg
from baldrapp.common import phasescreens
from baldrapp.common import pupils




################## 
# configure our zwfs 
grid_dict = {
    "D":1, # diameter of beam 
    "N" : 64, # number of pixels across pupil diameter
    "padding_factor" : 4, # how many pupil diameters fit into grid x axis
    # TOTAL NUMBER OF PIXELS = padding_factor * N 
    }

optics_dict = {
    "wvl0" :1.65e-6, # central wavelength (m) 
    "F_number": 21.2, # F number on phasemask
    "mask_diam": 1.06, # diameter of phaseshifting region in diffraction limit units (physical unit is mask_diam * 1.22 * F_number * lambda)
    "theta": 1.57079, # phaseshift of phasemask 
}

dm_dict = {
    "dm_model":"BMC-multi-3.5",
    "actuator_coupling_factor":0.7,# std of in actuator spacing of gaussian IF applied to each actuator. (e.g actuator_coupling_factor = 1 implies std of poke is 1 actuator across.)
    "dm_pitch":1,
    "dm_aoi":0, # angle of incidence of light on DM 
    "opd_per_cmd" : 3e-6, # peak opd applied at center of actuator per command unit (normalized between 0-1) 
    "flat_rmse" : 20e-9 # std (m) of flatness across Flat DM  
    }

grid_ns = SimpleNamespace(**grid_dict)
optics_ns = SimpleNamespace(**optics_dict)
dm_ns = SimpleNamespace(**dm_dict)

detector = (4,4) 

zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)



scrn = phasescreens.PhaseScreenVonKarman(nx_size= zwfs_ns.grid.N * zwfs_ns.grid.padding_factor , pixel_scale= zwfs_ns.grid.D / zwfs_ns.grid.N , r0=0.1, L0=12)

tel = 'UT'
vibs = 'none'
pup = pupils.pick_pupil(tel, dim = zwfs_ns.grid.pupil_mask.shape[0], diameter = zwfs_ns.grid.N)
phase_scaling_factor  = 0.3


fig_path = f'/home/benja/Documents/BALDR/figures/exploring_optical_gain/gif_{tel}_phSclF-{phase_scaling_factor}_vibs-{vibs}/'

if os.path.exists(fig_path) == False:
    os.makedirs(fig_path)
    
for i in range(100):
    print(i)
    for _ in range(zwfs_ns.grid.N//10):
        scrn.add_row()

    input_phase = phase_scaling_factor * pup * scrn.scrn # radians 

    strehl = np.exp(-np.var( input_phase[pup>0.3] ) )
    print( 'S=', strehl )

    psi_A =  pup * np.exp(1j *input_phase)
    Nb = 128
    no_loD = Nb//2
    psi_B = mft( psi_A, Na = input_phase.shape[0], Nb = 128, m = no_loD, cpix=False)
    #plt.imshow( np.abs( psi_B ) ); plt.show()


    x = np.linspace( -no_loD/2 , no_loD/2 , Nb)
    y = x.copy()
    X,Y = np.meshgrid(x,y) 

    phasemask_mask = X**2 + Y**2 <= (no_loD / 32 )**2
    #plt.imshow( phasemask_mask );plt.show()


    b = imft(phasemask_mask * psi_B, Nb = input_phase.shape[0], Na = Nb, m = no_loD, cpix=False)

    r1, r2, c1, c2 = (zwfs_ns.grid.padding_factor-1)//2 *zwfs_ns.grid.N  , (  zwfs_ns.grid.padding_factor-1) *zwfs_ns.grid.N , (zwfs_ns.grid.padding_factor-1)//2 *zwfs_ns.grid.N  , (zwfs_ns.grid.padding_factor-1) *zwfs_ns.grid.N 
    im_list = [ input_phase[ r1:r2, c1:c2  ] , abs(psi_B)**2, np.real(b),  np.imag(b), np.abs(b)]
    xlabel_list = ['' for _ in im_list]
    ylabel_list = ['' for _ in im_list]
    title_list = [f'input phase\n(Strehl={round(100* strehl)}%)', 'PSF', 'Real\noptical gain', 'Imaginary\noptical gain', 'Magnitude\noptical gain']
    cbar_label_list = ['radian', 'ADU', 'Re[b]', 'Im[b]', '|b|']
    util.nice_heatmap_subplots( im_list , xlabel_list, ylabel_list, title_list, cbar_label_list, fontsize=15,\
        cbar_orientation = 'bottom', axis_off=True, vlims=None, savefig=fig_path + f'frame_{i}.png')

    #plt.show()
    plt.close('all')
    

# psi_A2 = imft(psi_B, Nb = input_phase.shape[0], Na = 100, m = 50, cpix=False)
# plt.imshow( np.abs( psi_B ) ); plt.show()



b = bldr.get_b( phi = input_phase , phasemask = zwfs_ns.grid.phasemask_mask )

b = bldr.get_b_fresnel( phi = input_phase , phasemask = zwfs_ns.grid.phasemask_mask , wavelength=zwfs_ns.optics.wvl0, dx = zwfs_ns.grid.D / zwfs_ns.grid.N , z = zwfs_ns.grid.D**2 / zwfs_ns.optics.wvl0)

im_list = [ input_phase , np.real(b), np.imag(b), np.abs(b)]
xlabel_list = ['' for _ in im_list]
ylabel_list = ['' for _ in im_list]
title_list = ['input phase', 'Real', 'Imaginary', 'Magnitude']
cbar_label_list = ['radian','Re[b]', 'Im[b]', '|b|']
util.nice_heatmap_subplots( im_list , xlabel_list, ylabel_list, title_list, cbar_label_list, fontsize=15,\
    cbar_orientation = 'bottom', axis_off=True, vlims=None, savefig=None)








def _mft(array, Na, Nb, m, inverse=False, cpix=False):
    '''
    Performs the matrix Fourier transform of an array.
    
    Based on the formalism detailed in:
    
        Fast computation of Lyot-style coronagraph propagation
        Soummer, Pueyo, Sivaramakrishnan and Vanderbei
        2007, Optics Express, 15, 15935

    Parameters
    ----------
    array : array
        The input array
    
    Na : int
        Number of pixels in direct space
    
    Nb : int
        Number of pixels in Fourier space
    
    m : float
        Number of lambda/D elements in Fourier space
    
    inverse : bool, optional
        Control if the direct or inverse transform is performed. Default is `False`
    
    cpix : bool, optional
        Center the MFT on one pixel, instead of between 4 pixels. Default is 'False'
        
    Returns
    -------
    return_value : array
        MFT of the input array
    '''
        
    if (inverse is True):
        sign = 1
    else:
        sign = -1
    
    # For pixel centering
    offsetA = 0
    offsetB = 0
    
    # If centering between 4 pixels
    if not(cpix):
        offsetA = .5/Na
        offsetB = .5/Nb
    
    coeff = m / (Na * Nb)
    
    x = np.linspace(-0.5, 0.5, Na, endpoint=False, dtype=np.double) + offsetA
    y = x
    
    u = m * (np.linspace(-0.5, 0.5, Nb, endpoint=False, dtype=np.double) + offsetB)
    v = u
    
    A1 = np.exp(sign*2j*np.pi*np.outer(u, x))
    A3 = np.exp(sign*2j*np.pi*np.outer(v, y))
    
    B = coeff * A1.dot(array).dot(A3.T)

    return B


def mft(array, Na, Nb, m, cpix=False):
    '''
    Performs the matrix Fourier transform of an array.
    
    Based on the formalism detailed in:
    
        Fast computation of Lyot-style coronagraph propagation
        Soummer, Pueyo, Sivaramakrishnan and Vanderbei
        2007, Optics Express, 15, 15935

    Parameters
    ----------
    array : array
        The input array
    
    Na : int
        Number of pixels in direct space
    
    Nb : int
        Number of pixels in Fourier space
    
    m : float
        Number of lambda/D elements in Fourier space
    
    cpix : bool, optional
        Center the MFT on one pixel, instead of between 4 pixels. Default is 'False'
         
    Returns
    -------
    return_value : array
        MFT of the input array
    '''
    
    return _mft(array, Na, Nb, m, inverse=False, cpix=cpix)


def imft(array, Na, Nb, m, cpix=True):
    '''
    Performs the inverse matrix Fourier transform of an array.
    
    Based on the formalism detailed in:
    
        Fast computation of Lyot-style coronagraph propagation
        Soummer, Pueyo, Sivaramakrishnan and Vanderbei
        2007, Optics Express, 15, 15935

    Parameters
    ----------
    array : array
        The input array
    
    Na : int
        Number of pixels in direct space
    
    Nb : int
        Number of pixels in Fourier space
    
    m : float
        Number of lambda/D elements in Fourier space
        
    Returns
    -------
    return_value : array
        MFT of the input array
    '''
    
    return _mft(array, Na, Nb, m, inverse=True, cpix=cpix)
